@@ -17,7 +17,7 @@ from __future__ import absolute_import
 import hashlib
 import http.client
 
-import mock
+from unittest import mock
 import pytest  # type: ignore
 
 from google.resumable_media import _helpers
@@ -425,6 +425,31 @@ class Test__parse_generation_header(object):
         assert generation_header == self.GENERATION_VALUE
 
 
+class Test__is_decompressive_transcoding(object):
+    def test_empty_value(self):
+        headers = {}
+        response = _mock_response(headers=headers)
+        assert _helpers._is_decompressive_transcoding(response, _get_headers) is False
+
+    def test_gzip_in_headers(self):
+        headers = {_helpers._STORED_CONTENT_ENCODING_HEADER: "gzip"}
+        response = _mock_response(headers=headers)
+        assert _helpers._is_decompressive_transcoding(response, _get_headers) is True
+
+    def test_gzip_not_in_headers(self):
+        headers = {_helpers._STORED_CONTENT_ENCODING_HEADER: "identity"}
+        response = _mock_response(headers=headers)
+        assert _helpers._is_decompressive_transcoding(response, _get_headers) is False
+
+    def test_gzip_w_content_encoding_in_headers(self):
+        headers = {
+            _helpers._STORED_CONTENT_ENCODING_HEADER: "gzip",
+            _helpers.CONTENT_ENCODING_HEADER: "gzip",
+        }
+        response = _mock_response(headers=headers)
+        assert _helpers._is_decompressive_transcoding(response, _get_headers) is False
+
+
 class Test__get_generation_from_url(object):
 
     GENERATION_VALUE = 1641590104888641
@@ -468,6 +493,14 @@ class Test__add_query_parameters(object):
         assert new_url == "{}&{}".format(MEDIA_URL, expected)
 
 
+def test__get_uploaded_checksum_from_headers_error_handling():
+    response = _mock_response({})
+
+    with pytest.raises(ValueError):
+        _helpers._get_uploaded_checksum_from_headers(response, None, "invalid")
+    assert _helpers._get_uploaded_checksum_from_headers(response, None, None) is None
+
+
 def _mock_response(headers):
     return mock.Mock(
         headers=headers,
